// LICENSE : MIT
"use strict";
const Promise = require("bluebird");
const fs = require("fs");
const path = require("path");
const debug = require("debug")("textlint:cli");
const mkdirp = require("mkdirp");
import options from "./options";
import TextLintEngine from "./textlint-engine";
import TextFixEngine from "./textfix-engine";
import Config from "./config/config";
import configInit from "./config/config-initializer";
import TextLintFixer from "./fixer/textlint-fixer";
import Logger from "./util/logger";
/*
 cli.js is command line **interface**

 processing role is cli-engine.js.
 @see cli-engine.js
 */

/** @typedef {Object} TextLintFormatterOption
 *  @property {string} formatterName
 *  @property {boolean} noColor
 */

/**
 * Print results of lining text.
 * @param {string} output the output text which is formatted by {@link TextLintEngine.formatResults}
 * @param {object} options cli option object {@lint ./options.js}
 * @returns {boolean} does print result success?
 */
function printResults(output, options) {
    if (!output) {
        return true;
    }
    const outputFile = options.outputFile;
    if (outputFile) {
        const filePath = path.resolve(process.cwd(), outputFile);
        if (fs.existsSync(filePath) && fs.statSync(filePath).isDirectory()) {
            Logger.error("Cannot write to output file path, it is a directory: %s", outputFile);
            return false;
        }
        try {
            mkdirp.sync(path.dirname(filePath));
            fs.writeFileSync(filePath, output);
        } catch (ex) {
            Logger.error("There was a problem writing the output file:\n%s", ex);
            return false;
        }
    } else {
        Logger.log(output);
    }
    return true;
}

/**
 * Encapsulates all CLI behavior for eslint. Makes it easier to test as well as
 * for other Node.js programs to effectively run the CLI.
 */
const cli = {
    /**
     * Executes the CLI based on an array of arguments that is passed in.
     * @param {string|Array|Object} args The arguments to process.
     * @param {string} [text] The text to lint (used for TTY).
     * @returns {Promise<number>} The exit code for the operation.
     */
    execute(args, text) {
        var currentOptions;
        try {
            currentOptions = options.parse(args);
        } catch (error) {
            Logger.error(error.message);
            return Promise.resolve(1);
        }
        const files = currentOptions._;
        if (currentOptions.version) {
            // version from package.json
            Logger.log(`v${ require("../package.json").version }`);
        } else if (currentOptions.init) {
            return configInit.initializeConfig(process.cwd());
        } else if (currentOptions.help || !files.length && !text) {
            Logger.log(options.generateHelp());
        } else {
            // specify file name of stdin content
            const stdinFilename = currentOptions.stdinFilename;
            debug(`Running on ${ text ? "text" : "files" }, stdin-filename: ${stdinFilename}`);
            return this.executeWithOptions(currentOptions, files, text, stdinFilename);
        }
        return Promise.resolve(0);
    },
    /**
     * execute with cli options
     * @param {object} cliOptions
     * @param {string[]} files files are file path list
     * @param {string} text?
     * @param {string} stdinFilename?
     * @returns {Promise<number>} exit status
     */
    executeWithOptions(cliOptions, files, text, stdinFilename){
        const config = Config.initWithCLIOptions(cliOptions);
        const showEmptyRuleWarning = () => {
            Logger.log(`
== Not have rules, textlint do not anything ==
=> How to set rule?
See https://github.com/textlint/textlint/blob/master/docs/configuring.md
`);
        };

        if (cliOptions.fix) {
            // --fix
            const fixEngine = new TextFixEngine(config);
            if (!fixEngine.hasRuleAtLeastOne()) {
                showEmptyRuleWarning();
                return Promise.resolve(0);
            }
            const resultsPromise = text ? fixEngine.executeOnText(text, stdinFilename)
                : fixEngine.executeOnFiles(files);
            return resultsPromise.then(results => {
                const fixer = new TextLintFixer();
                const output = fixEngine.formatResults(results);
                printResults(output, cliOptions);
                // --dry-run
                if(cliOptions.dryRun){
                    debug("Enable dry-run mode.");
                    return Promise.resolve(0);
                }
                // modify file and return exit status
                return fixer.write(results).then(() => {
                    return 0;
                });
            });
        }
        // lint as default
        const lintEngine = new TextLintEngine(config);
        if (!lintEngine.hasRuleAtLeastOne()) {
            showEmptyRuleWarning();
            return Promise.resolve(0);
        }
        const resultsPromise = text ? lintEngine.executeOnText(text, stdinFilename) : lintEngine.executeOnFiles(files);
        return resultsPromise.then(results => {
            const output = lintEngine.formatResults(results);
            if (printResults(output, cliOptions)) {
                return lintEngine.isErrorResults(results) ? 1 : 0;
            } else {
                return 1;
            }
        });
    }
};
module.exports = cli;
